from typing import Optional

import requests

from ScanWatch.exceptions import APIException
from ScanWatch.utils.enums import NETWORK


class Client:
    """
    Client for the etherscan.io API or the bscscan.io API
    https://etherscan.io/apis
    https://bscscan.com/apis
    """
    BASE_URLS = {
        NETWORK.BSC: {
            "main": "https://api.bscscan.com/api",
            "test": "https://api-testnet.bscscan.com/api"
        },
        NETWORK.ETHER: {
            "main": "https://api.etherscan.io/api",
            "goerli": "https://api-goerli.etherscan.io/api",
            "kovan": "https://api-kovan.etherscan.io/api",
            "rinkeby": "https://api-rinkeby.etherscan.io/api",
            "ropsten": "https://api-ropsten.etherscan.io/api"
        }
    }

    def __init__(self, api_token: str, nt_type: NETWORK, net: str = "main"):
        """


        :param api_token: token for the api
        :type api_token: str
        :param nt_type: type of the network
        :type nt_type: NETWORK
        :param net: name of the network, used to differentiate main and test nets
        :type net: str, default 'main'
        """
        self.api_token = api_token
        self.nt_type = nt_type
        self.net = net
        self.get_url_request()  # test if network parameters are valid

    def get_mined_blocks(self, address: str, start_block: Optional[int] = None, end_block: Optional[int] = None):
        """
        fetch mined blocks by an address

        :param address: network address
        :type address: str
        :param start_block: fetch mined blocks starting with this block
        :type start_block: Optional[int]
        :param end_block: fetch mined blocks until this block
        :type end_block: Optional[int]
        :return: List of mined blocks
        :rtype: List[Dict]
        """
        try:
            return self._get_transactions(address, 'getminedblocks', start_block, end_block)
        except APIException:
            return []

    def get_erc721_transactions(self, address: str, start_block: Optional[int] = None, end_block: Optional[int] = None):
        """
        fetch erc721 transactions on an eth / bsc address

        :param address: address
        :type address: str
        :param start_block: fetch transactions starting with this block
        :type start_block: Optional[int]
        :param end_block: fetch transactions until this block
        :type end_block: Optional[int]
        :return: List of transactions
        :rtype: List[Dict]
        """
        return self._get_transactions(address, 'tokennfttx', start_block, end_block)

    def get_erc20_transactions(self, address: str, start_block: Optional[int] = None, end_block: Optional[int] = None):
        """
        fetch erc20 transactions on an eth / bsc address

        :param address: address
        :type address: str
        :param start_block: fetch transactions starting with this block
        :type start_block: Optional[int]
        :param end_block: fetch transactions until this block
        :type end_block: Optional[int]
        :return: List of transactions
        :rtype: List[Dict]

        .. code-block:: python

            [{'blockNumber': '108941',
              'timeStamp': '148216518',
              'hash': '0948461czecc9ze8e4vsvbq94sd96',
              'nonce': '23908745',
              'blockHash': '0x74a984dz56c13v8ze9q451vda',
              'from': 'zazd9f1wsqda84zds5qd6zda',
              'contractAddress': 'azd984f1edazdadqfefa',
              'to': '84cazd984csgzefa984zq5s1c',
              'value': '15248960000000000000',
              'tokenName': 'ChainLink Token',
              'tokenSymbol': 'LINK',
              'tokenDecimal': '18',
              'transactionIndex': '45',
              'gas': '200001',
              'gasPrice': '100000000000',
              'gasUsed': '51481',
              'cumulativeGasUsed': '1491504',
              'input': 'deprecated',
              'confirmations': '1948513'},
            ...
            ]

        """
        return self._get_transactions(address, 'tokentx', start_block, end_block)

    def get_normal_transactions(self, address: str, start_block: Optional[int] = None, end_block: Optional[int] = None):
        """
        fetch normal transactions on an eth / bsc address

        :param address: address
        :type address: str
        :param start_block: fetch transactions starting with this block
        :type start_block: Optional[int]
        :param end_block: fetch transactions until this block
        :type end_block: Optional[int]
        :return: List of transactions
        :rtype: List[Dict]

        .. code-block:: python

            [{'blockNumber': '10272495',
              'timeStamp': '1485965131',
              'hash': 'd9azfv1q9zf84zr15f49f49zef1z3sd1g98t1b6',
              'nonce': '9846651',
              'blockHash': 'zad94v16s1qef9a4f9v1r53b1sq64daf',
              'from': '496a1ef65s1dbv4a96z513svez965q',
              'contractAddress': 'az41f8ze1f63q5s1gv89ez49',
              'to': 'az49f84161ac89s4ef984a96e',
              'value': '17854000000000000000',
              'tokenName': 'ChainLink Token',
              'tokenSymbol': 'LINK',
              'tokenDecimal': '18',
              'transactionIndex': '79',
              'gas': '200001',
              'gasPrice': '100000000000',
              'gasUsed': '84215',
              'cumulativeGasUsed': '1531404',
              'input': 'deprecated',
              'confirmations': '119452'},
            ...
            ]
        """
        return self._get_transactions(address, 'txlist', start_block, end_block)

    def get_internal_transactions(self, address: str, start_block: Optional[int] = None,
                                  end_block: Optional[int] = None):
        """
        fetch internal transactions on an eth / bsc address

        :param address: address
        :type address: str
        :param start_block: fetch transactions starting with this block
        :type start_block: Optional[int]
        :param end_block: fetch transactions until this block
        :type end_block: Optional[int]
        :return: List of transactions
        :rtype: List[Dict]
        """
        return self._get_transactions(address, 'txlistinternal', start_block, end_block)

    def _get_transactions(self, address: str, action: str, start_block: Optional[int] = None,
                          end_block: Optional[int] = None):
        """
        fetch transactions on an eth / bsc address

        :param address: address
        :type address: str
        :param action: name of the request for the api (ex 'txlist' or 'txlistinternal')
        :type action:
        :param start_block: fetch transactions starting with this block
        :type start_block: Optional[int]
        :param end_block: fetch transactions until this block
        :type end_block: Optional[int]
        :return: List of transactions
        :rtype: List[Dict]
        """
        offset = 10000
        page_number = 1
        transactions = []
        while True:
            url = self.get_url_request(module='account',
                                       action=action,
                                       sort='asc',
                                       address=address,
                                       startblock=start_block,
                                       endblock=end_block,
                                       page=page_number,
                                       offset=offset)
            batch_txs = self.get_result(url)
            transactions.extend(batch_txs)
            if len(batch_txs) < offset:
                break
            else:
                page_number += 1
        return transactions

    def get_balance(self, address: str) -> float:
        """
        fetch the current eth / bnb balance of an address

        :param address: address
        :type address: str
        :return: ETH amount
        :rtype: float
        """
        url = self.get_url_request(module='account',
                                   action='balance',
                                   address=address,
                                   tag='latest'
                                   )
        return float(self.get_result(url))

    def get_url_request(self, **kwargs) -> str:
        """
        Construct the url to make a request to the etherscan.io / bscscan.com API

        :param kwargs: keywords args for the endpoint
        :type kwargs: Any
        :return:
        :rtype:
        """
        _keywords = {**kwargs, "apikey": self.api_token}
        string_kws = "&".join((f"{key}={value}" for key, value in _keywords.items()))
        try:
            base_url = self.BASE_URLS[self.nt_type][self.net]
        except KeyError as err:
            raise ValueError(f"unknown network with type {self.nt_type} and name {self.net}") from err
        return f"{base_url}?{string_kws}"

    @staticmethod
    def get_result(url: str):
        """
        call the API with an url, raise if the status is not ok and return the API result

        :param url: url to request for the etherscan.io or bscscan.com
        :type url: str
        :return: API result
        :rtype: depend of the endpoint
        """
        response = requests.get(url, headers={"User-Agent": "Mozilla/5.0"})
        response.raise_for_status()
        r_json = response.json()
        if int(r_json['status']) > 0 or r_json['message'] == 'No transactions found':
            return r_json['result']
        raise APIException(response)



from typing import Dict, List

from ScanWatch.storage.DataBase import DataBase
from ScanWatch.storage.tables import get_transaction_table
from ScanWatch.utils.enums import TRANSACTION, NETWORK


class ScanDataBase(DataBase):
    """
    Handles the recording of the address transactions in a local database
    """

    def __init__(self, name: str = 'scan_db'):
        """
        Initialise a Scan database instance

        :param name: name of the database
        :type name: str
        """
        super().__init__(name)

    def add_transactions(self, address: str, nt_type: NETWORK, net: str, tr_type: TRANSACTION, transactions: List[Dict]):
        """
        Add a list of transactions to the database

        :param address: address involved in the transaction
        :type address: str
        :param nt_type: type of network
        :type nt_type: NETWORK
        :param net: name of the network, used to differentiate main and test nets
        :type net: str
        :param tr_type: type of the transaction to record
        :type tr_type: TRANSACTION
        :param transactions: list of the transaction to record
        :type transactions: List[Dict]
        :return: None
        :rtype: None
        """
        table = get_transaction_table(address, nt_type, net, tr_type)
        for transaction in transactions:
            row = table.dict_to_tuple(transaction)
            self.add_row(table, row, auto_commit=False)
        self.commit()

    def get_transactions(self, address: str, nt_type: NETWORK, net: str, tr_type: TRANSACTION) -> List[Dict]:
        """
        Return the List of the transactions recorded in the database

        :param address: address involved in the transactions
        :type address: str
        :param nt_type: type of network
        :type nt_type: NETWORK
        :param net: name of the network, used to differentiate main and test nets
        :type net: str
        :param tr_type: type of the transaction to fetch
        :type tr_type: TRANSACTION
        :return: list of the transaction recorded
        :rtype: List[Dict]
        """
        table = get_transaction_table(address, nt_type, net, tr_type)
        rows = self.get_all_rows(table)
        return [table.tuple_to_dict(row) for row in rows]

    def get_last_block_number(self, address: str, nt_type: NETWORK, net: str, tr_type: TRANSACTION) -> int:
        """
        Return the last block number seen in recorded transactions (per address, type of transaction and network)
        If None are found, return 0

        :param address: address involved in the transactions
        :type address: str
        :param nt_type: type of network
        :type nt_type: NETWORK
        :param net: name of the network, used to differentiate main and test nets
        :type net: str
        :param tr_type: type of the transaction to fetch
        :type tr_type: TRANSACTION
        :return: last block number
        :rtype: int
        """
        table = get_transaction_table(address, nt_type, net, tr_type)
        selection = str(table.blockNumber)
        query = self.get_conditions_rows(table, selection=selection)
        default = 0
        try:
            return max([int(e[0]) for e in query])
        except ValueError:
            return default






class APIException(Exception):

    def __init__(self, response):
        self.code = 0
        try:
            json_res = response.json()
        except ValueError:
            self.message = 'Invalid JSON error message from the API: {}'.format(response.text)
        else:
            self.code = int(json_res['status'])
            self.message = json_res['result']
        self.status_code = response.status_code
        self.response = response
        self.request = getattr(response, 'request', None)

    def __str__(self):  # pragma: no cover
        return 'APIError(code=%s): %s' % (self.code, self.message)



from tqdm import tqdm

from ScanWatch.Client import Client
from ScanWatch.storage.ScanDataBase import ScanDataBase
from ScanWatch.utils.enums import NETWORK, TRANSACTION


class ScanManager:
    """
    This class is the interface between the user, the API and the Database
    """

    def __init__(self, address: str, nt_type: NETWORK, api_token: str, net: str = "main"):
        """
        Initiate the manager

        :param address: address to monitor
        :type address: str
        :param nt_type: type of the network
        :type nt_type: NETWORK
        :param api_token: token to communicate with the API
        :type api_token: str
        :param net: name of the network, used to differentiate main and test nets
        :type net: str, default 'main'
        """
        self.address = address
        self.nt_type = nt_type
        self.net = net
        self.client = Client(api_token, self.nt_type, self.net)
        self.db = ScanDataBase()

    def update_transactions(self, tr_type: TRANSACTION):
        """
        Update the transactions of a certain type in the database

        :param tr_type: type of transaction to update
        :type tr_type: TRANSACTION
        :return: None
        :rtype: None
        """
        last_block = self.db.get_last_block_number(self.address, self.nt_type, self.net, tr_type)
        if tr_type == TRANSACTION.NORMAL:
            new_transactions = self.client.get_normal_transactions(self.address, start_block=last_block + 1)
        elif tr_type == TRANSACTION.INTERNAL:
            new_transactions = self.client.get_internal_transactions(self.address, start_block=last_block + 1)
        elif tr_type == TRANSACTION.ERC20:
            new_transactions = self.client.get_erc20_transactions(self.address, start_block=last_block + 1)
        elif tr_type == TRANSACTION.ERC721:
            new_transactions = self.client.get_erc721_transactions(self.address, start_block=last_block + 1)
        else:
            raise ValueError(f"unknown transaction type: {tr_type}")
        self.db.add_transactions(self.address, self.nt_type, self.net, tr_type, new_transactions)

    def update_all_transactions(self):
        """
        Update all the transactions for the address

        :return: None
        :rtype: None
        """
        tr_types_names = [name for name in dir(TRANSACTION) if not name.startswith('__')]
        pbar = tqdm(total=len(tr_types_names))
        for name in tr_types_names:
            pbar.set_description(f"fetching {name.lower()} transactions for {self.nt_type.name.lower()} "
                                 f"address {self.address[:5]}...{self.address[-5:]}")
            self.update_transactions(getattr(TRANSACTION, name))
            pbar.update()
        pbar.set_description(f"all transactions updated for address {self.address[:5]}...{self.address[-5:]}")
        pbar.close()

    def get_transactions(self, tr_type: TRANSACTION):
        """
        Return the transactions of the provided type that are saved locally for the address of the manager

        :param tr_type: type of transaction to fetch
        :type tr_type: TRANSACTION
        :return: list of transactions
        :rtype: List[Dict]
        """
        return self.db.get_transactions(self.address, self.nt_type, self.net, tr_type)



import os
from setuptools import setup

this_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

about = {}
with open(os.path.join(this_directory, 'ScanWatch/__init__.py'), encoding='utf-8') as f:
    exec(f.read(), about)

with open("requirements.txt") as file:
    requirements_content = file.read()
requirements = requirements_content.split('\n')

setup(
    name='ScanWatch',
    version=about['__version__'],
    packages=['ScanWatch', 'ScanWatch.storage', 'ScanWatch.utils'],
    url='https://github.com/EtWnn/ScanWatch',
    author='EtWnn',
    author_email='EtWnn0x@gmail.com',
    license='MIT',
    description='Local tracker of an eth address for ETH and BSC scan',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    install_requires=requirements,
    keywords='eth bsc wallet save tracking history ethereum tracker binance smartchain smart chain',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



from typing import List, Optional, Tuple, Dict

from ScanWatch.utils.enums import NETWORK, TRANSACTION


class Table:
    """
    This class represent a table in a database. All columns names are dynamic attributes
    @DynamicAttrs
    This class is used to describe the tables that will be used to in the database
    """

    def __init__(self, name: str, columns_names: List[str], columns_sql_types: List[str],
                 primary_key: Optional[str] = None, primary_key_sql_type: Optional[str] = None):
        """
        Initialise a Table instance

        :param name: name of the table
        :type name: str
        :param columns_names: names of the columns (except primary column)
        :type columns_names: List[str]
        :param columns_sql_types: sql types of the previous columns
        :type columns_sql_types: List[str]
        :param primary_key: name of the primary key (None, if no primary key is needed)
        :type primary_key: Optional[str]
        :param primary_key_sql_type: sql type of the primary key (None, if no primary key is needed)
        :type primary_key_sql_type: Optional[str]
        """
        self.name = name
        self.columns_names = columns_names
        self.columns_sql_types = columns_sql_types
        self.primary_key = primary_key
        self.primary_key_sql_type = primary_key_sql_type

        for column_name in self.columns_names:
            try:
                value = getattr(self, column_name)
                raise ValueError(f"the name {column_name} conflicts with an existing attribute of value {value}")
            except AttributeError:
                setattr(self, column_name, column_name)

        if self.primary_key is not None:
            setattr(self, self.primary_key, self.primary_key)

    def tuple_to_dict(self, row: Tuple) -> Dict:
        """
        Transform a row from Tuple to Dict with column names as keys.

        :param row: a row of this table
        :type row: Tuple
        :return: the dictionary equivalent of this row
        :rtype: Dict
        """
        keys = []
        if self.primary_key is not None:
            keys.append(self.primary_key)
        keys += self.columns_names
        if len(keys) != len(row):
            raise ValueError(f"{len(keys)} values were expected but the row submitted only has {len(row)}")
        return {k: v for k, v in zip(keys, row)}

    def dict_to_tuple(self, row: Dict) -> Tuple:
        """
        Transform a row from Dict to Tuple, in the order of the table columns

        :param row: a row of this table
        :type row: Dict
        :return: the tuple equivalent of this row
        :rtype: Tuple
        """
        keys = []
        if self.primary_key is not None:
            keys.append(self.primary_key)
        keys += self.columns_names
        try:
            return tuple([row[k] for k in keys])
        except KeyError as err:
            raise ValueError(f"missing keys in the row provided: {keys} are expected") from err


def get_normal_transaction_table(address: str, scan_type: NETWORK):
    """
    Return the table used to store the normal transactions for an address and a scan type

    :param address: address of the transactions
    :type address: str
    :param scan_type: scan type: ether or bsc
    :type scan_type: str
    :return: table to store the transactions
    :rtype: Table
    """
    rows = [
        'blockNumber',
        'timeStamp',
        'hash',
        'nonce',
        'blockHash',
        'transactionIndex',
        'from',
        'to',
        'value',
        'gas',
        'gasPrice',
        'isError',
        'txreceipt_status',
        'input',
        'contractAddress',
        'cumulativeGasUsed',
        'gasUsed',
        'confirmations',
    ]
    row_types = len(rows) * ['TEXT']
    return Table(f"{scan_type}_{address}_normal_transaction", rows, row_types)


def get_transaction_table(address: str, nt_type: NETWORK, net: str, tr_type: TRANSACTION):
    """
    Return the table used to store the transactions depending on the address, network type and transaction type

    :param address: address of the transactions
    :type address: str
    :param nt_type: type of network
    :type nt_type: NETWORK
    :param net: name of the network, used to differentiate main and test nets
    :type net: str
    :param tr_type: type of the transaction to record
    :type tr_type: TRANSACTION
    :return: corresponding table
    :rtype: Table
    """
    if tr_type == TRANSACTION.NORMAL:
        rows = [
            'blockNumber',
            'timeStamp',
            'hash',
            'nonce',
            'blockHash',
            'transactionIndex',
            'from',
            'to',
            'value',
            'gas',
            'gasPrice',
            'isError',
            'txreceipt_status',
            'input',
            'contractAddress',
            'cumulativeGasUsed',
            'gasUsed',
            'confirmations',
        ]
    elif tr_type == TRANSACTION.ERC20:
        rows = [
            'blockNumber',
            'timeStamp',
            'hash',
            'nonce',
            'blockHash',
            'from',
            'contractAddress',
            'to',
            'value',
            'tokenName',
            'tokenSymbol',
            'tokenDecimal',
            'transactionIndex',
            'gas',
            'gasPrice',
            'gasUsed',
            'cumulativeGasUsed',
            'input',
            'confirmations',
        ]
    elif tr_type == TRANSACTION.ERC721:
        rows = [
            'blockNumber',
            'timeStamp',
            'hash',
            'nonce',
            'blockHash',
            'from',
            'contractAddress',
            'to',
            'tokenID',
            'tokenName',
            'tokenSymbol',
            'tokenDecimal',
            'transactionIndex',
            'gas',
            'gasPrice',
            'gasUsed',
            'cumulativeGasUsed',
            'input',
            'confirmations',
        ]
    elif tr_type == TRANSACTION.INTERNAL:
        rows = [
            'blockNumber',
            'timeStamp',
            'hash',
            'from',
            'to',
            'value',
            'contractAddress',
            'input',
            'type',
            'gas',
            'gasUsed',
            'traceId',
            'isError',
            'errCode'
        ]
    else:
        raise ValueError(f"unknown transaction type: {tr_type}")

    row_types = len(rows) * ['TEXT']
    pre_name = f"{nt_type.name.lower()}_{tr_type.name.lower()}"
    if net != "main":  # backward compatibility
        pre_name += f"_{net}"
    return Table(pre_name + f"_{address}_transaction", rows, row_types)
